#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2016 Truong Do <truongdq54@gmail.com>
#
# Distributed under terms of the MIT license.

"""

"""

try:
    from pynvml import *
except:
    print "Warning: pyvnml is not available, cannot detect GPU card automatically"

def get_gpu_available():
    gpu = -1 
    try:
        nvmlInit()
        deviceCount = nvmlDeviceGetCount()
        max_mem_free = -1
        
        for i in range(0, deviceCount):
            handle = nvmlDeviceGetHandleByIndex(i)
            memInfo = nvmlDeviceGetMemoryInfo(handle)
            mem_total = memInfo.total 
            mem_used = memInfo.used 
            mem_free = memInfo.free
            
            if mem_free > max_mem_free:
                max_mem_free = mem_free
                gpu = i
    finally:
        return gpu

if __name__ == "__main__":
    print get_gpu_available()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2016 truong-d <truong-d@ahclab40>
#
# Distributed under terms of the MIT license.

"""

"""
from __future__ import print_function
import chainer
import chainer.links as L
import chainer.functions as F
from chainer import ChainList

class StackLSTM(ChainList):
    def __init__(self, I, O, depth=1, drop_ratio=0):
        chain_list = []
        for i in range(depth):
            start = I if i == 0 else O
            chain_list.append(L.LSTM(start, O))
        self._drop_ratio = drop_ratio
        super(StackLSTM, self).__init__(*chain_list)
        self._train = True
    
    @property
    def train(self):
        return self._train
    
    @train.setter
    def train(self, value):
        self._train = value

    def reset_state(self):
        for lstm in self:
            lstm.reset_state()

    def __call__(self, inp, is_train=False):
        ret = None
        for i, hidden in enumerate(self):
            h = inp if i == 0 else ret
            ret = hidden(h)
        if self.train and self._drop_ratio:
            return F.dropout(ret, train=is_train, ratio=self._drop_ratio)
        else:
            return ret

    def get_state(self):
        ret = []
        for lstm in self:
            ret.append((lstm.c, lstm.h))
        return ret

    @property
    def h(self):
        return self.get_state()[-1][-1]

    @h.setter
    def h(self, h):
        for lstm_self in self:
            lstm_self.h = h

    def set_state(self, state):
        for lstm_self, lstm_in in zip(self, state):
            lstm_self.c, lstm_self.h = lstm_in


class BaseAttention(chainer.Chain):
    def __init__(self, osize):
        super(BaseAttention, self).__init__()
        self.osize = osize

    def __call__(self, enc_mat, h, output_weight=False):
        weights = F.softmax(F.batch_matmul(enc_mat, h))
        att = F.reshape(F.batch_matmul(weights, enc_mat, transa=True), (h.data.shape[0], self.osize))
        if not output_weight:
            return att
        else:
            return att, weights

    def reset(self):
        pass


class BaseDecoder(chainer.Chain):
    def __init__(self, isize, osize, depth=1, drop_ratio=0.):
        super(BaseDecoder, self).__init__(
                dec=StackLSTM(isize, osize, depth, drop_ratio)
             )
        self._train = True
    
    def train(self, value):
        self._train = value
        self.dec.train = value

    def reset(self):
        self.dec.reset_state()

    def __call__(self, x):
        return self.dec(x)


class BaseEncoder(chainer.Chain):
    def __init__(self, isize, osize, depth=1, drop_ratio=0.):
        super(BaseEncoder, self).__init__(
                encF=StackLSTM(isize, osize, depth, drop_ratio),
                encB=StackLSTM(isize, osize, depth, drop_ratio),
                aw = L.Linear(osize * 2, osize)
                )
        self._train = True
    
    def train(self, value):
        self._train = value
        self.encF.train = value
        self.encB.train = value

    def _encode_forward(self, x):
        return self.encF(x)

    def _encode_backward(self, x):
        return self.encB(x)

    def encode(self, x_list):
        self.reset()
        fx_list = []  # forward encoded list
        bx_list = []  # backward encoded list
        for idx, x in enumerate(x_list):
            fx = self._encode_forward(x)   # encode forward
            bx = self._encode_backward(x_list[-idx - 1])  # encode backward
            fx_list.append(fx)
            bx_list.append(bx)

        e_list = []
        for idx in xrange(len(x_list)):
            fx_i = fx_list[idx]
            bx_i = bx_list[-idx - 1]
            e_i = self.aw(F.concat((fx_i, bx_i), axis=1))
            e_list.append(e_i)
        return e_list

    def reset(self):
        self.encF.reset_state()
        self.encB.reset_state()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2016 truong-d <truong-d@ahclab40>
#
# Distributed under terms of the MIT license.

"""

"""
from __future__ import print_function
import chainer
import chainer.links as L
import chainer.functions as F
import base

class MTEncoder(base.BaseEncoder):
    def __init__(self, isize, osize, depth=1, drop_ratio=0.):
        super(MTEncoder, self).__init__(isize, osize, depth, drop_ratio)
        self._train = True
    
    @property
    def train(self):
        return self._train
    
    @train.setter
    def train(self, value):
        self._train = value
        super(MTEncoder, self).train(value)

    def __call__(self, x_list):
        batch_size = x_list[0].data.shape[0]
        e_list = super(MTEncoder, self).encode(x_list)
        S = F.reshape(F.concat(e_list, axis=1), (batch_size, len(x_list), -1))
        return S, e_list[-1]

    def reset(self):
        super(MTEncoder, self).reset()


class MTDecoder(base.BaseDecoder):
    def __init__(self, isize, osize, depth=1, drop_ratio=0.):
        super(MTDecoder, self).__init__(isize, isize, depth, drop_ratio)
        self.add_link('wc', L.Linear(isize * 2, isize))
        self.add_link('wo', L.Linear(isize, osize))
        self._train = True
    
    @property
    def train(self):
        return self._train
    
    @train.setter
    def train(self, value):
        self._train = value
        super(MTDecoder, self).train(value)

    def update(self, x):
        self.h = super(MTDecoder, self).__call__(x)
        return self.h

    def __call__(self, att_in=None, output_hidden=False):
        if att_in is not None:
            h = F.tanh(self.wc(F.concat((self.h, att_in), axis=1)))
        else:
            h = self.h
        self.o = self.wo(h)
        if output_hidden:
            return self.o, h
        else:
            return self.o

    def reset(self, e_l=None):
        super(MTDecoder, self).reset()
        if e_l is not None:
           self.start_with(e_l)

    def start_with(self, x):
        o = super(MTDecoder, self).__call__(x)


class MTAttention(base.BaseAttention):
    def __init__(self, osize):
        super(MTAttention, self).__init__(osize)

    def __call__(self, enc_mat, h, output_weight=False):
        return super(MTAttention, self).__call__(enc_mat, h, output_weight=output_weight)




#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2016 Truong Do <truongdq54@gmail.com>
#
# Distributed under terms of the MIT license.

"""

"""
import sys
import codecs
import subprocess
from contextlib import contextmanager

def read_from_pipe(program):
    cmd = program
    proc = subprocess.Popen([cmd], stdout=subprocess.PIPE, shell=True)
    return proc.stdout

def write_to_pipe(program):
    proc = subprocess.Popen([program], stdin=subprocess.PIPE, shell=True)
    return proc.stdin, proc

@contextmanager
def smart_open(fname=None, mode="r", encode=None):
    is_pipe = False
    is_pipe_write = False
    proc = None
    fd = None
    try:
        fd, proc, is_pipe_write, is_pipe = get_file_descriptor(fname, mode, encode)
        yield fd
    finally:
        if fd:
            if is_pipe_write and proc:
                proc.communicate()
            if not is_pipe:
                fd.close()

def get_file_descriptor(fname, mode, encode=None):
    fd = None
    is_pipe_write = False
    is_pipe = False
    proc = None
    if (not fname) or (fname == "-"):
        if "r" in mode:
            fd = sys.stdin
        else:
            fd = sys.stdout
    elif "|" == fname[-1]:
        is_pipe = True
        fname = fname[:-1]
        fd = read_from_pipe(fname)
    elif "|" == fname[0]:
        fname = fname[1:]
        is_pipe_write = True
        fd, proc = write_to_pipe(fname)
    else:
        if encode:
            fd = codecs.open(fname, mode, encode)
        else:
            fd = open(fname, mode)
    return fd, proc, is_pipe_write, is_pipe


class SmartFile(object):
    def __init__(self, fname=None, mode="r"):
        self.fd, self.proc, self.is_pipe_write, self.is_pipe = get_file_descriptor(fname, mode)
        self.name = fname

    def close(self):
        if self.fd:
            if self.is_pipe_write and self.proc:
                self.proc.communicate()
            if not self.is_pipe:
                self.fd.close()

    def write(self, data):
        self.fd.write(data)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2016 Truong Do <truongdq54@gmail.com>
#
# Distributed under terms of the MIT license.

"""

"""
from __future__ import print_function
from etra.io import smart_open
from collections import OrderedDict

def load_vocab(fname, encode=None):
    vocab = OrderedDict()

    with smart_open(fname, "r", encode=encode) as fin:
        for line in fin:
            w, idx = line.strip().split()
            idx = int(idx)
            vocab[w] = idx
    return vocab

def pad(data, val, max_len):
    for row in data:
        row_len = len(row)
        if row_len > max_len:
            print("error padding")
            exit(1)
        for i in range(max_len- row_len):
            row.append(val)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2016 truong-d <truong-d@ahclab40>
#
# Distributed under terms of the MIT license.

"""

"""
from __future__ import print_function
import chainer
import component as com
import chainer.functions as F
import chainer.links as L
from chainer import cuda
import numpy as np
import os
from chainer import serializers
import json
import ConfigParser
from chainer import optimizers

xp = np

class ATT_NMT(chainer.Chain):
    def __init__(self, src_vocab_size, trg_vocab_size, emb_size, hsize, depth=1, \
            drop_ratio=0.):
        print('~~~~~~~ Model parameters ~~~')
        print('~ Embed size:', emb_size)
        print('~ Hidden size:', hsize)
        print('~ Layer depth:', depth)
        print('~ Drop ratio:', drop_ratio)
        print('~~~~~~~~~~~~~~~~~~~~~~~~~~~')

        self.emb_size = emb_size
        self.src_vocab_size = src_vocab_size
        self.trg_vocab_size = trg_vocab_size
        self.hsize = hsize
        self.depth = depth
        self.drop_ratio = drop_ratio

        enc = com.MTEncoder(emb_size, hsize, depth, drop_ratio)
        att = com.MTAttention(hsize)
        dec = com.MTDecoder(hsize, hsize, depth, drop_ratio)
        h2o = L.Linear(hsize, trg_vocab_size)

        e2h = L.Linear(emb_size, hsize)
        src_emb = L.EmbedID(src_vocab_size, emb_size)
        trg_emb = L.EmbedID(trg_vocab_size, emb_size)

        super(ATT_NMT, self).__init__(enc=enc, att=att, dec=dec, e2h=e2h, src_emb=src_emb, trg_emb=trg_emb, h2o=h2o)
        self.S = None
        self._train = True
    
    @property
    def train(self):
        return self._train
    
    @train.setter
    def train(self, value):
        self._train = value
        self.enc.train = value
        self.dec.train = value

    def to_gpu(self):
        global xp
        xp = cuda.cupy
        super(ATT_NMT, self).to_gpu()

    def encode(self, x_batch):
        assert self.src_emb is not None
        x_list = []
        for i in xrange(x_batch.shape[1]):
            src_w = chainer.Variable(xp.asarray(x_batch[: , i], dtype=np.int32), volatile=not self.train)
            src_w = self.src_emb(src_w)
            x_list.append(src_w)

        self.S, e_l = self.enc(x_list)
        self.dec.reset(e_l)

    def reset(self):
        self.S = None
        self.enc.reset()
        self.dec.reset()
        self.att.reset()

    def decode(self, x, no_att=False, oweight=False):
        h = self.dec.update(self.e2h(x))
        weight = None
        if no_att:
            o = self.dec()
        else:
            att, weight = self.att(self.S, h, output_weight=True)
            o = self.dec(att_in=att)
        o = self.h2o(o)
        if oweight:
            return o, weight
        else:
            return o
    
    @classmethod
    def load(cls, model_dir):
        # Input files
        fn_model = os.path.join(model_dir, 'nnet.mdl')
        fn_cfg = os.path.join(model_dir, 'nnet.cfg')

        config = ConfigParser.RawConfigParser()
        config.read(fn_cfg)
        emb_size = config.getint('Structure', 'emb_size')
        src_vocab_size = config.getint('Structure', 'src_vocab_size')
        trg_vocab_size = config.getint('Structure', 'trg_vocab_size')
        hsize = config.getint('Structure', 'hsize')
        depth = config.getint('Structure', 'depth')
        drop_ratio = config.getfloat('Structure', 'drop_ratio')

        nnet = cls(src_vocab_size, trg_vocab_size, emb_size, hsize, depth, drop_ratio)
        serializers.load_npz(fn_model, nnet)

        return nnet

    def save(self, model_dir):
        if not os.path.exists(model_dir):
            os.makedirs(model_dir)

        config = ConfigParser.RawConfigParser()
        config.add_section('Structure')
        config.set('Structure', 'emb_size', str(self.emb_size))
        config.set('Structure', 'src_vocab_size', str(self.src_vocab_size))
        config.set('Structure', 'trg_vocab_size', str(self.trg_vocab_size))
        config.set('Structure', 'hsize', str(self.hsize))
        config.set('Structure', 'depth', str(self.depth))
        config.set('Structure', 'drop_ratio', str(self.drop_ratio))

        # Output files
        fn_model = os.path.join(model_dir, 'nnet.mdl')
        fn_cfg = os.path.join(model_dir, 'nnet.cfg')
        
        # Write out models
        serializers.save_npz(fn_model, self)
        
        with open(fn_cfg, 'w') as fd_cfg:
            config.write(fd_cfg)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2016 Truong Do <truongdq54@gmail.com>
#
# Distributed under terms of the MIT license.

"""

"""
from __future__ import print_function
import sys
import struct
import numpy as np
from os.path import abspath
import random
import tempfile
from etra.io import SmartFile, smart_open


def chunks(l, n):
    """Yield successive n-sized chunks from l."""
    for i in range(0, len(l), n):
        if i + n > len(l):
            yield l[i:i + n] + [l[-1] for x in range(i + n - len(l))]
        else:
            yield l[i:i + n]

class FeatWriter(object):
    def __init__(self, fname_scp=None, fname_ark=None):
        self.fd_scp = SmartFile(fname_scp, "w")
        if not fname_ark:
            fname_ark = fname_scp.replace(".scp", ".ark") if ".scp" in fname_scp else fname_scp + ".ark"
        self.fd_ark = SmartFile(fname_ark, "w")
        self.start = 0

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.fd_scp.close()
        self.fd_ark.close()

    def write(self, name, feat):
        nrow, ncol = feat.shape
        feat = feat.flatten()
        byte_size = feat.nbytes + 8  # 8 is header size
        if self.fd_scp:
            self.fd_scp.write("%s %s:%d:%d\n" % (name, abspath(self.fd_ark.name), self.start, byte_size))
        if self.fd_ark:
            self.fd_ark.write(struct.pack("<2i", nrow, ncol))
            self.fd_ark.write(feat)
        self.start += byte_size

class FeatReader(object):
    def __init__(self, fname=None, data_type=np.float32):
        self.name = fname
        self.data_type = data_type
        if fname:
            self.fd_scp = SmartFile(fname, "r")
        self.file_list = []

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, type, value, traceback):
        self.fd_scp.close()

    def open(self):
        self.file_list = self.fd_scp.fd.readlines()
        self.load()

    def read_all(self, shuffle=False):
        if shuffle:
            random.shuffle(self.file_list)
        for line in self.file_list:
            name, finfo = line.strip().split()
            fn_feat, start, length = finfo.split(':')
            start, length = int(start), int(length)
            yield name, self.read_ra((fn_feat, start, length))

    def size(self):
        return len(self.file_list)

    def read(self, name):
        return self.read_ra((self.fname_dict[name]))

    def read_ra(self, info):
        fn_feat, start, length = info
        header_size = 8
        length = int(length)
        with smart_open(fn_feat, 'rb') as fd_ark:
            fd_ark.seek(start)
            data = fd_ark.read(length)
            nrow, ncol = struct.unpack("<2i", data[0:header_size])
            feat = np.fromstring(data[header_size:], dtype=self.data_type)
            feat.shape = (nrow, ncol)
            return feat

    def copy(self):
        a = FeatReader()
        a.name = self.name
        a.file_list = self.file_list
        a.fname_list = self.fname_list
        a.fname_dict = self.fname_dict
        return a

    def load(self, reverse=True):
        self.fname_list = []
        for line in self.file_list:
            name, finfo = line.strip().split()
            fn_feat, start, length = finfo.split(':')
            length, start = int(length), int(start)
            self.fname_list.append((name, start, length, fn_feat))
        self.fname_dict = {name: (fn_feat, start, length) for name, start, length, fn_feat in self.fname_list}

    def get_list(self, bsize=1):
        return chunks(self.fname_list, bsize)

class RawFeatReader(object):
    def __init__(self, fname=None):
        self.fd = SmartFile(fname, "r")

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.fd.close()
        pass

    def read_all(self, dtype=np.float32):
        while not self.fd.fd.closed:
            header_size = 12
            header = self.fd.fd.read(header_size)
            if not header:
                break
            name_len, nrow, ncol = struct.unpack("<3i", header)
            item_size = 4  # both int and float has size of 4
            name = self.fd.fd.read(name_len)
            data = self.fd.fd.read(nrow * ncol * item_size)
            feat = np.fromstring(data, dtype=np.float32)
            feat.shape = (nrow, ncol)
            yield name, feat

class RawFeatWriter(object):
    def __init__(self, fname=None):
        self.fd_ark = SmartFile(fname, "w")
        self.start = 0

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        pass

    def write(self, name, feat):
        nrow, ncol = feat.shape
        feat = feat.flatten()
        self.fd_ark.write(struct.pack("<3i", len(name), nrow, ncol))
        self.fd_ark.write(name)
        self.fd_ark.write(feat)



"""
IndicoServer setup
For development:
    `python setup.py develop`
"""
from setuptools import setup, find_packages
import glob
import os

if __name__ == "__main__":
    setup(
        name="etra",
        packages = find_packages(), 
        install_requires=["setuptools", "nose", "coverage", "six",
            "chainer", "nvidia-ml-py"],
        version="0.1.2",
        author = 'Truong Do',
        author_email = 'truongdq54@gmail.com',
        url = 'https://github.com/truongdo/etra', # use the URL to the github repo
    )
